;;;; fp2scheme.scm


#+compiling
(declare
  (compress-literals 20)
  (export fp->scheme fp-parse fp-code))


(define-macro (tok t)
  `(make-token ,(list 'quasiquote t) yyline) )

(define-record-type token 
  (%make-token symbol value line)
  token?
  (symbol token-symbol)
  (value token-value)
  (line token-line) )

(define-record-printer (token x p)
  (fprintf p "#<token ~s:~a>" (token-value x) (token-line x)) )

(define (make-token t line)
  (cons 
   (car t)
   (match t
     ((sym val) (%make-token sym val line))
     ((sym) (%make-token sym #f line)) ) ) )

(define lexer-error error)

(define (escaped-char c)
  (case c
    ((#\n) #\newline)
    ((#\r) #\return)
    ((#\t) #\tab)
    ((#\e) #\esc)
    (else c) ) )

(include "fp.l.scm")

(define (parse-error msg #!optional arg)
  (match arg
    (#f (error msg))
    ((or (_ . ($ token symbol value line))
	 ($ token symbol value line) )
     (error (conc "(line " line ") " msg) symbol value) )
    (x (error msg x)) ) )

(include "grammar.lalr.scm")

(define (fp-parse s)
  (cond ((port? s) (lexer-init 'port s))
	((string? s) (lexer-init 'string s))
	(else (error 'fp-parse "bad argument type - not a string or port" s)) )
  (parser lexer parse-error) )

(define (fp->scheme top #!optional trace)
  (compile-toplevel top trace) )

(define *code-table* (make-hash-table eq?))
(define *pattern-bindings* '())

(define (fp-code sym)
  (hash-table-ref/default *code-table* sym #f))

(define (compile-toplevel x #!optional trace local)
  (define (defwrap x)
    (if trace
	`(fp:_trace _loc _arg (lambda () (,x _arg)))
	`(,x _arg)))
  (fluid-let ((*pattern-bindings* '()))
    (match x
      (('def ('once id) exp)
       (unless local (hash-table-set! *code-table* id exp))
       `(define ,(mangle id)
	  (let ((_f #f)
		(_loc ',id)
		(_val #f) )
	    (lambda (_arg)
	      (if _f
		  _val
		  (begin
		    (set! _val ,(defwrap (compile exp)))
		    (set! _f #t)
		    _val) ) ) ) ) )
      (('def id exp)
       (unless local (hash-table-set! *code-table* id exp))
       `(define ,(mangle id)
	  (let ((_loc ',id))
	    (lambda (_arg)
	      ,(defwrap (compile exp))))) )
      (('where ('def ('once id) exp) . body)
       (unless local (hash-table-set! *code-table* id exp))
       `(define ,(mangle id)
	  (let ((_f #f)
		(_loc ',id)
		(_val #f) )
	    (lambda (_arg)
	      ,@(map (cut compile-toplevel <> trace #t) body)
	      (if _f
		  _val
		  (begin
		    (set! _val ,(defwrap (compile exp)))
		    (set! _f #t)
		    _val) ) ) ) ) )
      (('where ('def id exp) . body)
       (unless local (hash-table-set! *code-table* id exp))
       `(define ,(mangle id)
	  (let ((_loc ',id))
	    (lambda (_arg)
	      ,@(map (cut compile-toplevel <> trace #t) body)
	      ,(defwrap (compile exp)) ) ) ) )
      (('app exp obj)
       `(let ((_loc '<TOPLEVEL>)
	      (_arg ,(compile-object obj)))
	  (,(compile exp) _arg) ) )
      (('seq ('atom files) ...)
       `(fp:_load ',files))
      (_ (error "invalid toplevel expression" x)) ) ) )

(define (mangle s)
  (string->symbol (conc "fp:" s)) )

(define (compile exp)
  (match exp
    (('cond x y z)
     (let* ((cx (compile x))
	    (cy (compile y))
	    (cz (compile z)))
       `(lambda (_x)
	  (if (eq? 'F (,cx _x)) 
	      (,cz _x) 
	      (,cy _x))) ) )
    (('comp f g)
     `(lambda (_x) (,(compile f) (,(compile g) _x))) )
    (('const o)
     `(lambda _ ,(compile-object o)) )
    (('cons fs ...)
     `(lambda (_x) (list ,@(map (lambda (f) `(,(compile f) _x)) fs))) )
    (('alpha f)
     `(lambda (_x) (map ,(compile f) _x)) )
    (('debug s)
     `(lambda (_x) (fp:_debug ',s _x) _x) )
    (('bottom msg)
     `(lambda (_x) (fp:_error _loc ',msg _arg)) )
    (('bu f o)
     `(lambda (_x) (,(compile f) (list _x ,(compile-object o)))))
    (('catch h f)
     `(lambda (_x) (fp:_catch ,(compile h) ,(compile f) _x)) )
    (('while p f)
     `(letrec ((loop 
		(lambda (_x)
		  (if (eq? 'F (,(compile p) _x))
		      _x
		      (loop (,(compile f) _x))))))
	loop) )
    (('select n)
     `(lambda (_x) (fp:_select ,n _x)))
    (('insert f)
     `(lambda (_x) (fp:_insert ,(compile f) _x)) )
    (('insert-right f)
     `(lambda (_x) (fp:_insert-right ,(compile f) _x)) )
    (('pcons fs ...)
     `(lambda (_x)
	(fp:_pcons
	 _x
	 ,@(let loop ((fs fs) (i 1))
	     (cond ((null? fs) '())
		   ((equal? '(...) fs) '('...))
		   ((match (car fs) (('let _ _) #t) (_ #f)) 
		    (set! *pattern-bindings* 
		      (alist-cons 
		       (cadar fs)
		       (if (eq? '... (caddar fs)) (- (sub1 i)) i)
		       *pattern-bindings*))
		    (loop (cons (caddar fs) (cdr fs)) i))
		   ((eq? '... (car fs)) 
		    (error "dots not at end of pattern" fs) )
		   (else (cons (compile (car fs)) (loop (cdr fs) (add1 i)))))))))
    (('verbatim a) (parse-scheme a))
    ((? symbol?)
     (cond ((assq exp *pattern-bindings*) => 
	    (lambda (a)
	      (if (zero? (cdr a))
		  (compile 'id)
		  (compile `(select ,(cdr a))))))
	   (else (mangle exp))))
    (_ (error "invalid expression" exp)) ) )

(define (compile-object o)
  (match o
    (('atom s) `',s)
    ('... '...)
    ((? number?) o)
    (('verbatim a) (parse-scheme a))
    (('seq xs ...) `(list ,@(map compile-object xs))) 
    (_ (error "invalid object" o)) ) )

(define (parse-scheme sym)
  (with-input-from-string (symbol->string sym) read) )

(include "fplib.scm")
