;;;; fplib.scm


(use srfi-1 utils regex posix)


#|

Functional forms:

 debug ATOM:x -> x
 error ATOM:x -> _|_
 /f:<x, ...> -> y
 \f:<x, ...> -> y
 {f, ...}:x -> F | <y, ...>

Primitives:

 show:x -> x
 eq:<x, y> -> bool
 apndl:<x, <y, ...>> -> <x, y, ...>
 apndr:<<x, ...>, y> -> <x, ..., y>
 +:<x, y> -> x + y
 -:<x, y> -> x - y
 x:<x, y> -> x * y
 %:<x, y> -> x / y
 gt:<x, y> -> x > y
 lt:<x, y> -> x < y
 id:x -> x
 type:x -> atom
 tl:<x1 x2, ...> -> <x2, ...>
 null:x -> bool
 length:<x1, ...> -> N
 reverse:<x1, ...> -> <..., x1>
 cat:<<x1, ...>, <y1, ...>> -> <x1, ..., y1, ...>
 atom:x -> bool
 code:a -> x
 forget:<a, ...> -> <a, ...>
 compile:<a, x> -> a
 pack:<n, ...> -> a
 unpack:x -> <n, ...>
 number:a -> n
 defined:_ -> <a, ...>
 read:n -> <c, ...>
 write:x
 throw:x
 system:a
 readf:a -> <c, ...>
 writef:<a, <c, ...>>
 existsf:a -> bool
 readp:a -> <c, ...>
 writep:<a, <c, ...>>
 statf:a -> <n, ...>

;|#


#+compiling
(declare 
  (export fp:_debug fp:_error fp:show fp:_insert
	  fp:_pcons fp:eq fp:apndl fp:apndr fp:+ fp:code
	  fp:- fp:x fp:% fp:gt fp:lt fp:id fp:tl fp:null
	  fp:compile fp:forget fp:pack fp:unpack 
	  fp:number fp:defined fp:_insert-right fp:read
	  fp:write fp:_catch fp:throw fp:_select
	  fp:system fp:_trace fp:_load fp:readf fp:writef
	  fp:existsf fp:readp fp:writep fp:statf
	  fp:reverse fp:length fp:cat fp:type fp:atom) )


(define-macro (binary name . body)
  `(match-lambda
     ((x y) ,@body)
     (args (error ',name "argument is not a sequence" args)) ) )

(define (boolify x)
  (if (eq? 'F x) #f #t) )

(define (unboolify x)
  (if x 'T 'F) )

(define (num loc pos x)
  (cond ((number? x) x)
	((symbol? x) 
	 (or (string->number (symbol->string x))
	     (badarg loc pos x "a number")))
	((char? x) (char->integer x))
	((string? x) 
	 (or (string->number x)
	     (badarg loc pos x "a number")))
	(else (badarg loc pos x "a number"))))

(define (atm loc pos x)
  (cond ((symbol? x) x)
	((number? x) (string->symbol (number->string x)))
	((string? x) (string->symbol x))
	((boolean? x) (if x 'T 'F))
	((char? x) (string->symbol (string x)))
	(else (badarg loc pos x "an atom"))))

(define (seq loc pos x)
  (cond ((pair? x) x)
	((null? x) x)
	((vector? x) (vector->list x))
	(else (badarg loc pos x "a sequence"))))

(define (badarg loc pos x msg)
  (error 
   loc
   (string-append
    "argument is not " 
    msg
    (if pos
	(sprintf " in position ~a" pos)
	"") )
   x) )

(define (fp:_select i x)
  (cond ((zero? i) '())
	((negative? i) (list-tail (seq i #f x) (- i)))
	(else (list-ref (seq i #f x) (sub1 i)))))

(define (fp:_debug s x)
  (fprintf (current-error-port) "~a: " s)
  (fp:show x) 
  x)

(define (fp:_error loc msg arg)
  (error (cond ((number? loc) (string->symbol (number->string loc)))
	       (else loc))
	 (atm '_ #f msg) arg))

(define fp:show
  (let ((rx (regexp "^[A-Z][A-Za-z0-9_]*$")))
    (lambda (x)
      (define (show1 x)
	(cond ((null? x) (display "<>"))
	      ((symbol? x) 
	       (let ((name (symbol->string x)))
		 ((if (string-match rx name) display write) name) ) )
	      ((atom? x) (display x))
	      (else
	       (write-char #\<)
	       (show1 (car x))
	       (do ((x (cdr x) (cdr x)))
		   ((null? x))
		 (display ", ")
		 (show1 (car x)) )
	       (write-char #\>) ) ) )
      (show1 x)
      (newline)
      x) ) )

(define (unit f)
  (cond ((eq? f fp:+) 0)
	((eq? f fp:x) 1)
	(else (error 'unit "unknown unit value" f))))

(define (fp:_insert f x)
  (if (null? x)
      (unit f)
      (reduce (lambda (x y) (f (list y x))) #f x) ) )

(define (fp:_insert-right f x)
  (if (null? x)
      (unit f)
      (reduce-right (lambda (x y) (f (list x y))) #f x) ) )

(define (fp:_pcons x . fs)
  (let loop ((x x) (fs fs))
    (cond ((null? fs) (unboolify (null? x)))
	  ((eq? '... (car fs)) 'T)
	  ((not (pair? x)) 'F)
	  ((boolify ((car fs) (car x)))
	   (loop (cdr x) (cdr fs)) )
	  (else 'F) ) ) )

(define fp:eq (binary eq (unboolify (equal? x y)) ))
(define fp:apndl (binary apndl (cons x (seq 'apndl 2 y))))
(define fp:apndr (binary apdnr (append (seq 'apndr 1 x) (list y))))
(define fp:+ (binary + (+ (num '+ 1 x) (num '+ 2 y))))
(define fp:- (binary - (- (num '- 1 x) (num '- 2 y))))
(define fp:x (binary x (* (num 'x 1 x) (num 'x 2 y))))
(define fp:% (binary % (quotient (num '% 1 x) (num '% 2 y))))
(define fp:gt (binary gt (unboolify (> (num 'gt 1 x) (num 'gt 2 y)))))
(define fp:lt (binary lt (unboolify (< (num 'lt 1 x) (num 'lt 2 y)))))
(define fp:id identity)
(define fp:tl cdr)
(define (fp:null x) (unboolify (null? x)))
(define (fp:reverse x) (reverse (seq 'reverse #f x)))
(define (fp:length x) (length (seq 'length #f x)))

(define (fp:cat x)
  (concatenate (map (cut seq 'cat #f <>) (seq 'cat #f x))))

(define (fp:type x)
  (cond ((symbol? x) 'ATOM)
	((number? x) 'NUMBER)
	((or (null? x) (pair? x)) 'SEQUENCE)
	(else (error 'type "internal - unknown type of object" x)) ) )

(define (fp:atom x)
  (unboolify (or (null? x) (symbol? x) (number? x))))

(define (fp:code x)
  (or (fp-code (atm 'code #f x)) 'F) )

(define (fp:forget lst)
  (if (pair? lst)
      (for-each (cut hash-table-delete! *code-table* <>) lst)
      (set! *code-table* (make-hash-table eq?)))
  lst)

(define (fp:compile x)
  (eval 
   (fp->scheme 
    `(def ,(atm 'compile 1 (car x)) 
	  ,(cadr (seq 'compile 2 (cadr x)))))))

(define (fp:_load files)
  (do ((i 0 (add1 i))
       (fs (seq '<load> #f files) (cdr fs)))
      ((null? fs) files)
    (eval
     `(begin 
	,@(map fp->scheme
	       (call-with-input-file (symbol->string (atm '<load> i (car fs)))
		 fp-parse)))) ) )

(define (fp:pack x)
  (string->symbol (list->string (map integer->char (seq 'pack #f x)))))

(define (fp:unpack x)
  (map char->integer (string->list (symbol->string (atm 'unpack #f x)))))

(define (fp:number x)
  (cond ((number? x) x)
	((symbol? x) (or (string->number (symbol->string x)) 'F))
	((char? x) (char->integer x))
	((string? x) (or (string->number x) 'F))
	(else 'F)))

(define (fp:defined _) (hash-table-keys *code-table*))

(define (fp:read n)
  (map char->integer
       (string->list
	(cond ((number? n) (read-string n))
	      ((eq? 'LINE n) (read-line))
	      (else (read-all) ) ) ) ) )

(define (fp:write x)
  (cond ((number? x) (write-char (integer->char x)))
	((pair? x) (for-each (o write-char integer->char) x))
	((or (symbol? x) (string? x) (char? x)) (display x))
	(else (badarg 'write #f x "a number or sequence")))
  x)

(define fp:_catch
  (let ((msg (condition-property-accessor 'exn 'message)))
    (lambda (h f x)
      (condition-case (f x)
	(ex (exn) (h (cons 'error (msg ex))))
	(ex () (h ex)) ) ) ) )

(define fp:throw signal)

(define (fp:system x)
  (system* (symbol->string (atm 'system #f x))))

(define *indent* "")

(define (fp:_trace loc arg thunk)
  (let ((indent #f)
	(result '_))
    (dynamic-wind
	(lambda ()
	  (printf "[~a~a:~s]~%" *indent* loc arg)
	  (set! indent *indent*)
	  (set! *indent* (string-append *indent* " ")) )
	(lambda ()
	  (set! result (thunk))
	  result)
	(lambda ()
	  (set! *indent* indent) 
	  (printf "[~a~a -> ~s]~%" *indent* loc result) ) ) ) )

(define (fp:readf fn)
  (map char->integer (string->list (read-all (symbol->string (atm 'readf #f fn))))))

(define (fp:writef x)
  (with-output-to-file (symbol->string (atm 'writef 1 (car x)))
    (lambda () (display (map integer->char (seq 'writef 2 (cadr x)))))))

(define (fp:existsf fn)
  (unboolify (file-exists? (atm 'existsf #f fn))))

(define (fp:statf fn)
  (file-stat (symbol->string (atm 'statf #f fn)) ))

(define (fp:readp cmd)
  (with-input-from-pipe (symbol->string (atm 'readp #f cmd))
    (lambda ()
      (map char->integer (string->list (read-all) ) ) ) ) )

(define (fp:writep x)
  (with-output-to-pipe 
   (symbol->string (atm 'writep 1 (car x)))
   (lambda ()
     (write-string
      (let ((arg (cadr x)))
	(cond ((string? arg) arg)
	      ((symbol? arg) (symbol->string arg))
	      ((number? arg) (number->string arg))
	      (else (list->string (map integer->char arg)))))))) )
